@@ -12,14 +12,25 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-"""Minimal Python S2 cell ID, S2 token and lat/lon conversion library."""
+"""Minimal Python S2 Geometry cell ID, token and lat/lon conversion library."""
 
 import math
-from typing import Tuple, Union
+import re
+from typing import Optional, Tuple
 
-import numpy as np
+__version__ = '1.4.0'
 
-__version__ = '1.1.1'
+
+#
+# s2cell exceptions
+#
+
+class InvalidCellID(Exception):
+    """Exception type for invalid cell IDs."""
+
+
+class InvalidToken(Exception):
+    """Exception type for invalid tokens."""
 
 
 #
@@ -28,29 +39,29 @@ __version__ = '1.1.1'
 
 # The maximum level supported within an S2 cell ID. Each level is represented by two bits in the
 # final cell ID
-_S2_MAX_LEVEL = np.uint32(30)
+_S2_MAX_LEVEL = 30
 
 # The maximum value within the I and J bits of an S2 cell ID
-_S2_MAX_SIZE = np.uint32(1) << _S2_MAX_LEVEL
+_S2_MAX_SIZE = 1 << _S2_MAX_LEVEL
 
 # The number of bits in a S2 cell ID used for specifying the base face
-_S2_FACE_BITS = np.uint32(3)
+_S2_FACE_BITS = 3
 
 # The number of bits in a S2 cell ID used for specifying the position along the Hilbert curve
-_S2_POS_BITS = np.uint32(2 * _S2_MAX_LEVEL + 1)
+_S2_POS_BITS = 2 * _S2_MAX_LEVEL + 1
 
 # The maximum value of the Si/Ti integers used when mapping from IJ to ST. This is twice the max
-# value of I and J, since Si/Ti allow referencing both the centre and edge of a leaf cell
-_S2_MAX_SI_TI = np.uint32(1) << (_S2_MAX_LEVEL + 1)
+# value of I and J, since Si/Ti allow referencing both the center and edge of a leaf cell
+_S2_MAX_SI_TI = 1 << (_S2_MAX_LEVEL + 1)
 
 # Mask that specifies the swap orientation bit for the Hilbert curve
-_S2_SWAP_MASK = np.uint64(1)
+_S2_SWAP_MASK = 1
 
 # Mask that specifies the invert orientation bit for the Hilbert curve
-_S2_INVERT_MASK = np.uint64(2)
+_S2_INVERT_MASK = 2
 
 # The number of bits per I and J in the lookup tables
-_S2_LOOKUP_BITS = np.uint32(4)
+_S2_LOOKUP_BITS = 4
 
 # Lookup table for mapping 10 bits of IJ + orientation to 10 bits of Hilbert curve position +
 # orientation. Populated later by _s2_init_lookups
@@ -62,18 +73,16 @@ _S2_LOOKUP_IJ = None
 
 # Lookup table of two bits of IJ from two bits of curve position, based also on the current curve
 # orientation from the swap and invert bits
-_S2_POS_TO_IJ = np.array([
+_S2_POS_TO_IJ = [
     [0, 1, 3, 2],  # 0: Normal order, no swap or invert
     [0, 2, 3, 1],  # 1: Swap bit set, swap I and J bits
     [3, 2, 0, 1],  # 2: Invert bit set, invert bits
     [3, 1, 0, 2],  # 3: Swap and invert bits set
-], dtype=np.uint64)
+]
 
 # Lookup for the orientation update mask of one of the four sub-cells within a higher level cell.
 # This mask is XOR'ed with the current orientation to get the sub-cell orientation.
-_S2_POS_TO_ORIENTATION_MASK = np.array([
-    _S2_SWAP_MASK, 0, 0, _S2_SWAP_MASK | _S2_INVERT_MASK
-], dtype=np.uint64)
+_S2_POS_TO_ORIENTATION_MASK = [_S2_SWAP_MASK, 0, 0, _S2_SWAP_MASK | _S2_INVERT_MASK]
 
 
 #
@@ -112,7 +121,7 @@ def _s2_st_to_uv(component: float) -> float:
     return (1.0 / 3.0) * (1.0 - 4.0 * (1.0 - component) ** 2)
 
 
-def _s2_st_to_ij(component: float) -> np.uint64:
+def _s2_st_to_ij(component: float) -> int:
     """
     Convert S2 ST to IJ.
 
@@ -122,10 +131,12 @@ def _s2_st_to_ij(component: float) -> np.uint64:
     See s2geometry/blob/2c02e21040e0b82aa5719e96033d02b8ce7c0eff/src/s2/s2coords.h#L333-L336
 
     """
-    return np.uint64(max(0, min(_S2_MAX_SIZE - 1, round(_S2_MAX_SIZE * component - 0.5))))
+    # The reference implementation does round(_S2_MAX_SIZE * component - 0.5), which is equivalent
+    # to math.floor(_S2_MAX_SIZE * component)
+    return int(max(0, min(_S2_MAX_SIZE - 1, math.floor(_S2_MAX_SIZE * component))))
 
 
-def _s2_si_ti_to_st(component: np.uint64) -> float:
+def _s2_si_ti_to_st(component: int) -> float:
     """
     Convert S2 Si/Ti to ST.
 
@@ -138,7 +149,7 @@ def _s2_si_ti_to_st(component: np.uint64) -> float:
 
 
 def _s2_face_uv_to_xyz(  # pylint: disable=invalid-name
-        face: Union[int, np.uint64], uv: Tuple[float, float]
+        face: int, uv: Tuple[float, float]
 ) -> Tuple[float, float, float]:
     """
     Convert face + UV to S2Point XYZ.
@@ -176,7 +187,7 @@ def _s2_face_uv_to_xyz(  # pylint: disable=invalid-name
     elif face == 5:
         s2_point = (uv[1], uv[0], -1)
     else:
-        raise ValueError('Cannot convert UV to XYZ with invalid face: ' + str(face))
+        raise ValueError('Cannot convert UV to XYZ with invalid face: {}'.format(face))
 
     return s2_point
 
@@ -198,41 +209,43 @@ def _s2_init_lookups() -> None:
     """
     global _S2_LOOKUP_POS, _S2_LOOKUP_IJ  # pylint: disable=global-statement
     if _S2_LOOKUP_POS is None or _S2_LOOKUP_IJ is None:  # pragma: no branch
-        lookup_length = 1 << (2 * _S2_LOOKUP_BITS + 2)
-        _S2_LOOKUP_POS = np.zeros((lookup_length,), dtype=np.uint64)
-        _S2_LOOKUP_IJ = np.zeros((lookup_length,), dtype=np.uint64)
+        # Initialise empty lookup tables
+        lookup_length = 1 << int(2 * _S2_LOOKUP_BITS + 2)  # = 1024
+        _S2_LOOKUP_POS = [0] * lookup_length
+        _S2_LOOKUP_IJ = [0] * lookup_length
 
         # Generate lookups for each of the base orientations given by the swap and invert bits
-        for base_orientation in np.array([
-            0, _S2_SWAP_MASK, _S2_INVERT_MASK, _S2_SWAP_MASK | _S2_INVERT_MASK  # 0-4 effectively
-        ], dtype=np.uint64):
-            # Walk the 256 possible positions within a level 4 curve. There is probably a smarter
-            # way of doing this iteratively that reuses work at each level. e.g. 4 nested for loops?
-            for pos in np.arange(4 ** 4, dtype=np.uint64):
-                ij = np.uint64(0)  # Has pattern iiiijjjj, not ijijijij  # pylint: disable=invalid-name
-                orientation = np.uint64(base_orientation)
+        for base_orientation in [
+            0, _S2_SWAP_MASK, _S2_INVERT_MASK, _S2_SWAP_MASK | _S2_INVERT_MASK  # 0-3 effectively
+        ]:
+            # Walk the 256 possible positions within a level 4 curve. This implementation is not
+            # the fastest since it does not reuse the common ancestor of neighbouring positions, but
+            # is simpler to read
+            for pos in range(4 ** 4):  # 4 levels of sub-divisions
+                ij = 0  # Has pattern iiiijjjj, not ijijijij  # pylint: disable=invalid-name
+                orientation = base_orientation
 
-                # Walk the pairs of bits of pos, from most significant, getting IJ and orientation
-                # as we go
+                # Walk the pairs of bits of pos, from most significant to least, getting IJ and
+                # orientation as we go
                 for bit_pair_offset in range(4):
                     # Bit pair is effectively the sub-cell index
-                    bit_pair = (pos >> np.uint64((3 - bit_pair_offset) * 2)) & np.uint64(0b11)
+                    bit_pair = (pos >> ((3 - bit_pair_offset) * 2)) & 0b11
 
                     # Get the I and J for the sub-cell index. These need to be spread into iiiijjjj
                     # by inserting as bit positions 4 and 0
                     ij_bits = _S2_POS_TO_IJ[orientation][bit_pair]
                     ij = (  # pylint: disable=invalid-name
-                        (ij << np.uint64(1))  # Free up position 4 and 0 from old IJ
-                        | ((ij_bits & np.uint32(2)) << np.uint32(3))  # I bit in position 4
-                        | (ij_bits & np.uint32(1))  # J bit in position 0
+                        (ij << 1)  # Free up position 4 and 0 from old IJ
+                        | ((ij_bits & 2) << 3)  # I bit in position 4
+                        | (ij_bits & 1)  # J bit in position 0
                     )
 
                     # Update the orientation with the new sub-cell orientation
                     orientation = orientation ^ _S2_POS_TO_ORIENTATION_MASK[bit_pair]
 
                 # Shift IJ and position to allow orientation bits in LSBs of lookup
-                ij <<= np.uint32(2)  # pylint: disable=invalid-name
-                pos <<= np.uint32(2)
+                ij <<= 2  # pylint: disable=invalid-name
+                pos <<= 2
 
                 # Write lookups
                 _S2_LOOKUP_POS[ij | base_orientation] = pos | orientation
@@ -243,7 +256,7 @@ def _s2_init_lookups() -> None:
 # Cell ID <-> Token translation functions
 #
 
-def cell_id_to_token(cell_id: Union[int, np.uint64]) -> str:
+def cell_id_to_token(cell_id: int) -> str:
     """
     Convert S2 cell ID to a S2 token.
 
@@ -259,25 +272,22 @@ def cell_id_to_token(cell_id: Union[int, np.uint64]) -> str:
         The S2 token string for the S2 cell ID.
 
     Raises:
-        TypeError: If the cell_id is not int or np.uint64.
+        TypeError: If the cell_id is not int.
 
     """
     # Check type
-    if not isinstance(cell_id, (int, np.uint64)):
-        raise TypeError('Cannot convert S2 cell ID from type: ' + str(type(cell_id)))
-    cell_id = np.uint64(cell_id)
+    if not isinstance(cell_id, int):
+        raise TypeError('Cannot convert S2 cell ID from type: {}'.format(type(cell_id)))
 
-    # The zero token is encoded as 'X' rather than as a zero-length string. This implementation has
-    # no method of generating the 0 cell ID, so this line is mostly here for consistency with the
-    # reference implementation.
-    if cell_id == 0:  # pragma: no cover
+    # The zero token is encoded as 'X' rather than as a zero-length string
+    if cell_id == 0:
         return 'X'
 
-    # Convert cell ID to hex and strip any trailing zeros
+    # Convert cell ID to 16 character hex string and strip any implicit trailing zeros
     return '{:016x}'.format(cell_id).rstrip('0')
 
 
-def token_to_cell_id(token: str) -> np.uint64:
+def token_to_cell_id(token: str) -> int:
     """
     Convert S2 token to S2 cell ID.
 
@@ -293,25 +303,26 @@ def token_to_cell_id(token: str) -> np.uint64:
 
     Raises:
         TypeError: If the token is not str.
-        ValueError: If the token length is over 16.
+        InvalidToken: If the token length is over 16.
 
     """
     # Check input
     if not isinstance(token, str):
-        raise TypeError('Cannot convert S2 token from type: ' + str(type(token)))
-
-    if token in ('x', 'X'):
-        # The zero cell ID is represented as the character 'X' rather than as an empty string
-        return np.uint64(0)
+        raise TypeError('Cannot convert S2 token from type: {}'.format(type(token)))
 
     if len(token) > 16:
-        raise ValueError('Cannot convert S2 token with length > 16 characters')
+        raise InvalidToken('Cannot convert S2 token with length > 16 characters')
 
-    # Add stripped zeros
+    # Check for the zero cell ID represented by the character 'x' or 'X' rather than as the empty
+    # string
+    if token in ('x', 'X'):
+        return 0
+
+    # Add stripped implicit zeros to create the full 16 character hex string
     token = token + ('0' * (16 - len(token)))
 
     # Convert to cell ID by converting hex to int
-    return np.uint64(int(token, 16))
+    return int(token, 16)
 
 
 #
@@ -320,7 +331,7 @@ def token_to_cell_id(token: str) -> np.uint64:
 
 def lat_lon_to_cell_id(  # pylint: disable=too-many-locals
         lat: float, lon: float, level: int = 30
-) -> np.uint64:
+) -> int:
     """
     Convert lat/lon to a S2 cell ID.
 
@@ -377,7 +388,7 @@ def lat_lon_to_cell_id(  # pylint: disable=too-many-locals
     # -x -> 3
     # -y -> 4
     # -z -> 5
-    face = int(np.argmax(np.abs(s2_point)))  # Largest absolute component
+    face = max(enumerate(s2_point), key=lambda p: abs(p[1]))[0]  # Largest absolute component
     if s2_point[face] < 0.0:
         face += 3
 
@@ -438,30 +449,30 @@ def lat_lon_to_cell_id(  # pylint: disable=too-many-locals
     # levels, so the required number of steps is ceil((level + 2) / 4), when level is > 0. The
     # additional 2 levels added are required to account for the top 3 bits (4 before right shift)
     # that are occupied by the face bits.
-    bits = np.uint64(face) & _S2_SWAP_MASK  # iiiijjjjoo. Initially set by by face
-    cell_id = np.uint64(face << (_S2_POS_BITS - 1))  # Insert face at most signficant bits
-    lookup_mask = np.uint64((1 << _S2_LOOKUP_BITS) - 1)
+    bits = face & _S2_SWAP_MASK  # iiiijjjjoo. Initially set by by face
+    cell_id = face << (_S2_POS_BITS - 1)  # Insert face at second most signficant bits
+    lookup_mask = (1 << int(_S2_LOOKUP_BITS)) - 1  # Mask of 4 one bits: 0b1111
     required_steps = math.ceil((level + 2) / 4) if level > 0 else 0
     for k in range(7, 7 - required_steps, -1):
         # Grab 4 bits of each of I and J
-        offset = np.uint32(k * _S2_LOOKUP_BITS)
-        bits += ((ij[0] >> offset) & lookup_mask) << np.uint32(_S2_LOOKUP_BITS + 2)
-        bits += ((ij[1] >> offset) & lookup_mask) << np.uint32(2)
+        offset = k * _S2_LOOKUP_BITS
+        bits += ((ij[0] >> offset) & lookup_mask) << (_S2_LOOKUP_BITS + 2)
+        bits += ((ij[1] >> offset) & lookup_mask) << 2
 
-        # Map bits from iiiijjjjoo to ppppppppoo
+        # Map bits from iiiijjjjoo to ppppppppoo using lookup table
         bits = _S2_LOOKUP_POS[bits]
 
-        # Insert position bits to cell ID
-        cell_id |= (bits >> np.uint32(2)) << np.uint32(k * 2 * _S2_LOOKUP_BITS)
+        # Insert position bits into cell ID
+        cell_id |= (bits >> 2) << (k * 2 * _S2_LOOKUP_BITS)
 
         # Remove position bits, leaving just new swap and invert bits for the next round
-        bits &= (_S2_SWAP_MASK | _S2_INVERT_MASK)
+        bits &= _S2_SWAP_MASK | _S2_INVERT_MASK  # Mask: 0b11
 
     # Left shift and add trailing bit
     # The trailing bit addition is disabled, as we are overwriting this below in the truncation
     # anyway. This line is kept as an example of the full method for S2 cell ID creation as is done
     # in the standard library versions.
-    cell_id = (cell_id << np.uint8(1))  # + np.uint64(1)
+    cell_id = cell_id << 1  # + 1
 
     # Truncate to desired level
     # This is done by finding the mask of the trailing 1 bit for the specified level, then zeroing
@@ -469,7 +480,7 @@ def lat_lon_to_cell_id(  # pylint: disable=too-many-locals
     # still necessary to do even after a reduced number of steps `required_steps` above, since each
     # step contains multiple levels that may need partial overwrite. Additionally, we need to add
     # the trailing 1 bit, which is not yet set above.
-    least_significant_bit_mask = np.uint64(1) << np.uint32(2 * (_S2_MAX_LEVEL - level))
+    least_significant_bit_mask = 1 << (2 * (_S2_MAX_LEVEL - level))
     cell_id = (cell_id & -least_significant_bit_mask) | least_significant_bit_mask
 
     return cell_id
@@ -507,7 +518,7 @@ def lat_lon_to_token(lat: float, lon: float, level: int = 30) -> str:
 #
 
 def cell_id_to_lat_lon(  # pylint: disable=too-many-locals
-    cell_id: Union[int, np.uint64]
+    cell_id: int
 ) -> Tuple[float, float]:
     """
     Convert S2 cell ID to lat/lon.
@@ -519,14 +530,13 @@ def cell_id_to_lat_lon(  # pylint: disable=too-many-locals
         The lat/lon (in degrees) tuple generated from the S2 cell ID.
 
     Raises:
-        TypeError: If the cell_id is not int or np.uint64.
-        ValueError: If the cell_id has invalid face bits.
+        TypeError: If the cell_id is not int.
+        InvalidCellID: If the cell_id is invalid.
 
     """
-    # Check type
-    if not isinstance(cell_id, (int, np.uint64)):
-        raise TypeError('Cannot decode S2 cell ID from type: ' + str(type(cell_id)))
-    cell_id = np.uint64(cell_id)
+    # Check input
+    if not cell_id_is_valid(cell_id):
+        raise InvalidCellID('Cannot decode invalid S2 cell ID: {}'.format(cell_id))
 
     # Populate _S2_LOOKUP_IJ on first run.
     # See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2cell_id.cc#L75-L109
@@ -549,63 +559,68 @@ def cell_id_to_lat_lon(  # pylint: disable=too-many-locals
     # out, since these are not set in the IJ to cell ID during encoding.
     #
     # The I and J returned here are of one of the two leaf (level 30) cells that are located
-    # diagonally closest to the cell centre. This happens because repeated ..00.. will select the
+    # diagonally closest to the cell center. This happens because repeated ..00.. will select the
     # 'lower left' (for nominally oriented Hilbert curve segments) of the sub-cells. The ..10..
     # arising from the trailing bit, prior to the repeated ..00.. bits, ensures we first pick the
-    # 'upper right' of the cell, then iterate in to lower left until we hit the leaf cell. However,
-    # in the case of the swapped and inverted curve segment (4th sub-curve segment), the ..10.. will
-    # select the 'lower left' and then iterate to the 'upper right' with each ..00.. following. In
-    # that case, we will be offset left and down by one leaf cell in each of I and J, which needs to
-    # be added to have a consistent mapping. This is detectable by seeing that the final bit of I or
-    # J is 1 (i.e we have picked an odd row/column, which will happen concurrently in both I and J,
-    # so we only need to check one), except in case of level 29 where the logic is inverted and the
-    # correction needs to be applied when we pick an even row/column (i.e I/J ends in 0), since
-    # there are no trailing ..00..
+    # 'upper right' of the cell, then iterate in to lower left until we hit the leaf cell. This
+    # means we pick the leaf cell to the north east of the parent cell center (again for nominal
+    # orientation).
+    # However, in the case of the swapped and inverted curve segment (4th sub-curve segment), the
+    # ..10.. will select the 'lower left' and then iterate to the 'upper right' with each ..00..
+    # following. In that case, we will be offset left and down by one leaf cell in each of I and J,
+    # which needs to be fixed to have a consistent mapping. This is detectable by seeing that the
+    # final bit of I or J is 1 (i.e we have picked an odd row/column, which will happen concurrently
+    # in both I and J, so we only need to check one), except in case of level 29 where the logic is
+    # inverted and the correction needs to be applied when we pick an even row/column (i.e I/J ends
+    # in 0), since there are no trailing ..00..  available after the ``..10..`` when we are at level
+    # 29+.
     #
     # This behaviour can be captured in the expression:
     # apply_correction = not leaf and (i ^ (is level 29)) & 1
-    # apply_corerction = not leaf and (i ^ (cell_id >> 2)) & 1
+    # apply_correction = not leaf and (i ^ (cell_id >> 2)) & 1
     #
-    # We check for level 29 by looking for the trailing 1 in third LSB, when we already know that we
-    # are not a leaf cell (which could give false positive) by the initial check in the expression.
+    # We check for level 29 by looking for the trailing 1 in the third LSB, when we already know
+    # that we are not a leaf cell (which could give false positive) by the initial check in the
+    # expression.
     # See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2cell_id.h#L503-L529
     #
     face = cell_id >> _S2_POS_BITS
     bits = face & _S2_SWAP_MASK  # ppppppppoo. Initially set by by face
-    lookup_mask = np.uint64((1 << _S2_LOOKUP_BITS) - 1)
-    i = np.uint64(0)
-    j = np.uint64(0)
+    lookup_mask = (1 << _S2_LOOKUP_BITS) - 1  # Mask of 4 one bits: 0b1111
+    i = 0
+    j = 0
     for k in range(7, -1, -1):
         # Pull out 8 bits of cell ID, except in first loop where we pull out only 4
         n_bits = (_S2_MAX_LEVEL - 7 * _S2_LOOKUP_BITS) if k == 7 else _S2_LOOKUP_BITS
-        extract_mask = np.uint64((1 << (2 * n_bits)) - 1)  # 8 (or 4) one bits
+        extract_mask = (1 << (2 * n_bits)) - 1  # 8 (or 4) one bits
         bits += (
-            (cell_id >> np.uint32(k * 2 * _S2_LOOKUP_BITS + 1)) & extract_mask
-        ) << np.uint32(2)
+            (cell_id >> (k * 2 * _S2_LOOKUP_BITS + 1)) & extract_mask
+        ) << 2
 
-        # Map bits from ppppppppoo to iiiijjjjoo
+        # Map bits from ppppppppoo to iiiijjjjoo using lookup table
         bits = _S2_LOOKUP_IJ[bits]
 
         # Extract I and J bits
-        offset = np.uint32(k * _S2_LOOKUP_BITS)
-        i += (bits >> np.uint32(_S2_LOOKUP_BITS + 2)) << offset  # Don't need lookup mask here
-        j += ((bits >> np.uint32(2)) & lookup_mask) << offset
+        offset = k * _S2_LOOKUP_BITS
+        i += (bits >> (_S2_LOOKUP_BITS + 2)) << offset  # Don't need lookup mask here
+        j += ((bits >> 2) & lookup_mask) << offset
 
         # Remove I and J bits, leaving just new swap and invert bits for the next round
-        bits &= (_S2_SWAP_MASK | _S2_INVERT_MASK)
+        bits &= _S2_SWAP_MASK | _S2_INVERT_MASK  # Mask: 0b11
 
-    # Resolve the centre of the cell. For leaf cells, we add half the leaf cell size. For non-leaf
-    # cells, we currently have one of either two cells diagonally around the cell centre, as
-    # described above. The centre_correction_delta is 2x the offset, as we left shift I and J first.
+    # Resolve the center of the cell. For leaf cells, we add half the leaf cell size. For non-leaf
+    # cells, we currently have one of either two cells diagonally around the cell center and want
+    # to pick the leaf-cell edges that represent the parent cell center, as described above. The
+    # center_correction_delta is 2x the offset, as we left shift I and J first.
     # This gives us the values Si and Ti, which are discrete representation of S and T in range 0 to
-    # _S2_MAX_SI_TI. The extra power of 2 over IJ allows for identifying both the centre and edge of
+    # _S2_MAX_SI_TI. The extra power of 2 over IJ allows for identifying both the center and edge of
     # cells, whilst IJ is just the leaf cells.
     # See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2coords.h#L57-L65
-    is_leaf = bool(cell_id & np.uint64(1))  # Cell is leaf cell when trailing one bit is in LSB
-    apply_correction = not is_leaf and ((i ^ (cell_id >> np.uint32(2))) & np.uint64(1))
-    centre_correction_delta = np.uint64(1 if is_leaf else (2 if apply_correction else 0))
-    si = (i << np.uint32(1)) + centre_correction_delta  # pylint: disable=invalid-name
-    ti = (j << np.uint32(1)) + centre_correction_delta  # pylint: disable=invalid-name
+    is_leaf = bool(cell_id & 1)  # Cell is leaf cell when trailing one bit is in LSB
+    apply_correction = not is_leaf and ((i ^ (cell_id >> 2)) & 1)
+    correction_delta = 1 if is_leaf else (2 if apply_correction else 0)
+    si = (i << 1) + correction_delta  # pylint: disable=invalid-name
+    ti = (j << 1) + correction_delta  # pylint: disable=invalid-name
 
     # Convert integer si/ti to double ST
     # See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2coords.h#L338-L341
@@ -650,19 +665,135 @@ def token_to_lat_lon(token: str) -> Tuple[float, float]:
 
     Raises:
         TypeError: If the token is not str.
-        ValueError: If the token length is over 16.
-        ValueError: If the token has invalid face bits.
+        InvalidToken: If the token length is over 16.
+        InvalidToken: If the token is invalid.
+        InvalidCellID: If the contained cell_id is invalid.
 
     """
+    # Check input
+    if not token_is_valid(token):
+        raise InvalidToken('Cannot decode invalid S2 token: {}'.format(token))
+
     # Convert to cell ID and decode to lat/lon
     return cell_id_to_lat_lon(token_to_cell_id(token))
 
 
 #
-# Level extraction
+# Token canonicalisation
 #
 
-def cell_id_to_level(cell_id: Union[int, np.uint64]) -> int:
+def token_to_canonical_token(token: str) -> str:
+    """
+    Convert S2 token to a canonicalised S2 token.
+
+    This produces a token that matches the form generated by the reference C++ implementation:
+
+    - Lower case (except 'X' below)
+    - No whitespace
+    - Trailing '0' characters stripped
+    - Zero cell ID represented as 'X', not 'x' or ''
+
+    Args:
+        token: The S2 token string to canonicalise.
+
+    Returns:
+        The canonicalised S2 token.
+
+    """
+    # Convert token to lower case.
+    # Note that 'X' below will be returned upper case
+    token = token.lower()
+
+    # Strip any surrounding whitespace
+    token = token.strip()
+
+    # Strip any trailing zeros
+    token = token.rstrip('0')
+
+    # If empty string or 'x', return 'X' token
+    if token in ('', 'x'):
+        token = 'X'
+
+    return token
+
+
+#
+# Validation
+#
+
+def cell_id_is_valid(cell_id: int) -> bool:
+    """
+    Check that a S2 cell ID is valid.
+
+    Looks for valid face bits and a trailing 1 bit in one of the correct locations.
+
+    Args:
+        cell_id: The S2 cell integer to validate.
+
+    Returns:
+        True if the cell ID is valid, False otherwise.
+
+    Raises:
+        TypeError: If the cell_id is not int.
+
+    """
+    # Check input
+    if not isinstance(cell_id, int):
+        raise TypeError('Cannot decode S2 cell ID from type: {}'.format(type(cell_id)))
+
+    # Check for zero ID
+    # This avoids overflow warnings below when 1 gets added to max uint64
+    if cell_id == 0:
+        return False
+
+    # Check face bits
+    if (cell_id >> _S2_POS_BITS) > 5:
+        return False
+
+    # Check trailing 1 bit is in one of the even bit positions allowed for the 30 levels, using the
+    # mask: 0b0001010101010101010101010101010101010101010101010101010101010101 = 0x1555555555555555
+    lowest_set_bit = cell_id & (~cell_id + 1)  # pylint: disable=invalid-unary-operand-type
+    if not lowest_set_bit & 0x1555555555555555:
+        return False
+
+    return True  # Checks have passed, cell ID must be valid
+
+
+def token_is_valid(token: str) -> bool:
+    """
+    Check that a S2 token is valid.
+
+    Looks for valid characters, then checks that the contained S2 cell ID is also valid. Note that
+    the '', 'x' and 'X' tokens are considered invalid, since the cell IDs they represent are
+    invalid.
+
+    Args:
+        token: The S2 token string to validate.
+
+    Returns:
+        True if the token is valid, False otherwise.
+
+    Raises:
+        TypeError: If the token is not str.
+
+    """
+    # Check input
+    if not isinstance(token, str):
+        raise TypeError('Cannot check S2 token with type: {}'.format(type(token)))
+
+    # First check string with regex
+    if not re.match(r'^[0-9a-fA-f]{1,16}$', token):
+        return False
+
+    # Check the contained cell ID is valid
+    return cell_id_is_valid(token_to_cell_id(token))
+
+
+#
+# Level extraction functions
+#
+
+def cell_id_to_level(cell_id: int) -> int:
     """
     Get the level for a S2 cell ID.
 
@@ -675,18 +806,13 @@ def cell_id_to_level(cell_id: Union[int, np.uint64]) -> int:
         The level of the S2 cell ID.
 
     Raises:
-        TypeError: If the cell_id is not int or np.uint64.
-        ValueError: If the cell_id is 0.
+        TypeError: If the cell_id is not int.
+        InvalidCellID: If the cell_id is invalid.
 
     """
     # Check input
-    if not isinstance(cell_id, (int, np.uint64)):
-        raise TypeError('Cannot decode S2 cell ID from type: ' + str(type(cell_id)))
-    cell_id = np.uint64(cell_id)
-
-    if cell_id == 0:
-        # The cell ID 0 has no level
-        raise ValueError('Cannot decode invalid S2 cell ID: 0')
+    if not cell_id_is_valid(cell_id):
+        raise InvalidCellID('Cannot decode invalid S2 cell ID: {}'.format(cell_id))
 
     # Find the position of the lowest set one bit, which will be the trailing one bit. The level is
     # given by the max level (30) minus the floored division by two of the position of the lowest
@@ -696,10 +822,10 @@ def cell_id_to_level(cell_id: Union[int, np.uint64]) -> int:
     # equivalent to the C++20 function std::countr_zero() or the ctz instruction.
     lsb_pos = 0
     while cell_id != 0:  # pragma: no branch
-        if cell_id & np.uint64(1):
+        if cell_id & 1:
             break
         lsb_pos += 1
-        cell_id >>= np.uint64(1)
+        cell_id >>= 1
 
     return int(_S2_MAX_LEVEL - (lsb_pos >> 1))
 
@@ -718,9 +844,106 @@ def token_to_level(token: str) -> int:
 
     Raises:
         TypeError: If the token is not str.
-        ValueError: If the token length is over 16.
-        ValueError: If the token as cell ID is 0.
+        InvalidToken: If the token length is over 16.
+        InvalidToken: If the token is invalid.
+        InvalidCellID: If the contained cell_id is invalid.
 
     """
+    # Check input
+    if not token_is_valid(token):
+        raise InvalidToken('Cannot decode invalid S2 token: {}'.format(token))
+
     # Convert to cell ID and get the level for that
     return cell_id_to_level(token_to_cell_id(token))
+
+
+#
+# Parent functions
+#
+
+def cell_id_to_parent_cell_id(
+        cell_id: int, level: Optional[int] = None
+) -> int:
+    """
+    Get the parent cell ID of a S2 cell ID.
+
+    Args:
+        cell_id: The S2 cell ID integer.
+        level: The parent level to get the cell ID for. Must be less than or equal to the current
+            level of the provided cell ID. If unspecified, or None, the direct parent cell ID will
+            be returned.
+
+    Returns:
+        The parent cell ID at the specified level.
+
+    Raises:
+        TypeError: If the cell_id is not int.
+        InvalidCellID: If the cell_id is invalid.
+        ValueError: If cell ID is already level 0 and level is None.
+        ValueError: When level is not an integer, is < 0 or is > 30.
+        ValueError: If level is greater than the provided cell ID level.
+
+    """
+    # Check input
+    if not cell_id_is_valid(cell_id):
+        raise InvalidCellID('Cannot decode invalid S2 cell ID: {}'.format(cell_id))
+
+    # Get current level of the cell ID and check it is suitable with the requested level
+    current_level = cell_id_to_level(cell_id)
+    if level is None and current_level == 0:
+        raise ValueError('Cannot get parent cell ID of a level 0 cell ID')
+    if level is None:
+        level = current_level - 1
+
+    if not isinstance(level, int) or level < 0 or level > _S2_MAX_LEVEL:
+        raise ValueError('S2 level must be integer >= 0 and <= 30')
+
+    if level > current_level:
+        raise ValueError('Cannot get level {} parent cell ID of cell ID with level {}'.format(
+            level, current_level
+        ))
+    if level == current_level:
+        # Requested parent level is current level, return cell ID itself
+        return cell_id
+
+    # Truncate to desired level
+    # This is done by finding the mask of the trailing 1 bit for the specified level, then zeroing
+    # out all bits less significant than this, then finally setting the trailing 1 bit. This is
+    # still necessary to do even after a reduced number of steps `required_steps` above, since each
+    # step contains multiple levels that may need partial overwrite. Additionally, we need to add
+    # the trailing 1 bit, which is not yet set above.
+    least_significant_bit_mask = 1 << (2 * (_S2_MAX_LEVEL - level))
+    cell_id = (cell_id & -least_significant_bit_mask) | least_significant_bit_mask
+
+    return cell_id
+
+
+def token_to_parent_token(token: str, level: Optional[int] = None) -> str:
+    """
+    Get the parent token of a S2 token.
+
+    Args:
+        token: The S2 token string. Can be upper or lower case hex string.
+        level: The parent level to get the token for. Must be less than or equal to the current
+            level of the provided toke. If unspecified, or None, the direct parent token will be
+            returned.
+
+    Returns:
+        The parent token at the specified level.
+
+    Raises:
+        TypeError: If the token is not str.
+        InvalidToken: If the token length is over 16.
+        InvalidToken: If the token is invalid.
+        InvalidCellID: If the contained cell_id is invalid.
+        ValueError: If token is already level 0 and level is None.
+        ValueError: When level is not an integer, is < 0 or is > 30.
+        ValueError: If level is greater than the provided token level.
+
+    """
+    # Check input
+    if not token_is_valid(token):
+        raise InvalidToken('Cannot decode invalid S2 token: {}'.format(token))
+
+    # Convert to cell ID and get parent and convert back to token
+    return cell_id_to_token(cell_id_to_parent_cell_id(token_to_cell_id(token), level))
